"""
Pour simuler des robotsomes...
Règle adoptée : 2 robots sont bloqués s'ils convoitent la même case
Les autres cas (la file indienne ou les croisements) ne posent pas
de problèmes

C'est en POO, c'est un gros brouillon, c'est donc bien améliorable
C'est prévu pour tourner dans un notebook : ça affiche l'arène et
les déplacements en ASCII art dans la même cellule du notebook

Il faudrait y greffer une propriété type turtle pour un affichage graphique

Auteur : Sébastien Hoarau
Date   : 2021-09

Un exemple de grille (> est un robot qui regarde vers la droite
< regarde vers le gauche, V vers le bas et A vers le haut, il n'y 
en a pas sur cette grille)

##############
#.......>.<..#
#...#.....#..#
#...#..#.....#
#V...........#
# ############

"""

from IPython.display import clear_output
from time import sleep

# ---------
# CONSTANTS

# -- about type of cell content (wall or empty space)

WALL = 0
EMPTY = 1
EXIT = 2
MARKS = '#. '
CELL_MARK_TO_TYPE = {'#':WALL, '.':EMPTY, ' ':EXIT, 'A':EMPTY, 'V':EMPTY, '>':EMPTY, '<':EMPTY}

# -- about directions and dx,dy

NORTH = 0, -1
SOUTH = 0, 1
EAST = 1, 0
WEST = -1, 0

DIR_TO_MARK = {NORTH:'A', SOUTH:'V', EAST:'>', WEST:'<'}
MARK_TO_DIR = {'A':NORTH, 'V':SOUTH, '>':EAST, '<':WEST}
LEFT = {NORTH:WEST, WEST:SOUTH, SOUTH:EAST, EAST:NORTH}
RIGHT = {NORTH:EAST, EAST:SOUTH, SOUTH:WEST, WEST:NORTH}

# -- about default names of files and directories

REP_GRID = 'Grilles'


# -------
# CLASSES
        
class Bot:
    
    def __init__(self, bot_id, x, y, direction):
        self.__id = bot_id
        self.__x, self.__y = x, y
        self.__dx, self.__dy = direction
        self.__blocked = False
    
    @property
    def id(self):
        return self.__id

    @property
    def blocked(self):
        return self.__blocked
    
    @blocked.setter
    def blocked(self, value):
        self.__blocked = bool(value)

    @property
    def direction(self):
        return self.__dx, self.__dy
    
    @direction.setter
    def direction(self, new_direction):
        self.__dx, self.__dy = new_direction

    @property
    def position(self):
        return self.__x, self.__y
    
    @position.setter
    def position(self, new_position):
        self.__x, self.__y = new_position

    @property
    def destination(self):
        return self.__x + self.__dx, self.__y + self.__dy
    
    def debug(self):
        d = bot.direction
        print(f'bot {bot.id}: {d} ({DIR_TO_MARK[d]})')
    
    def __str__(self):
        return DIR_TO_MARK[self.direction]
    
    
    # -- about moving

    def left(self):
        self.direction = LEFT[self.direction]
        self.blocked = False

    def right(self):
        self.direction = RIGHT[self.direction]
        self.blocked = False

    def forward(self):
        if not self.blocked:
            self.position = self.destination
        self.blocked = False

        
class Arena:
    
    def __init__(self):
        self.w, self.h = 0, 0
        self.cells = {}
        self.bots = {}
        self.destinations = {}
        self.updated = False
        self.exit_coords = None
        
    def load(self, filename):
        current_id = 0
        with open(f'{REP_GRID}/{filename}.txt', 'r') as data:
            for y, line in enumerate(data):
                for x, car in enumerate(line.strip()):
                    car = car.upper()
                    self.cells[x,y] = CELL_MARK_TO_TYPE[car]
                    if car == MARKS[EXIT]:
                        self.exit_coords = x, y
                    elif car in MARK_TO_DIR:
                        self.bots[x,y] = Bot(current_id, x, y, MARK_TO_DIR[car])
                        current_id += 1
            self.w = x+1
            self.h = y+1

    def _mark(self, x, y, debug=False):
        if debug:
            return str(self.bot(x,y).id) if self.bot_at(x,y) else MARKS[self.cell(x,y)]
        else:
            return DIR_TO_MARK[self.bot(x,y).direction] if self.bot_at(x,y) else MARKS[self.cell(x,y)]
        
    def __str__(self):
        return '\n'.join([''.join([self._mark(x, y) for x in range(self.w)]) for y in range(self.h)])

    def debug(self):
        print('\n'.join([''.join([self._mark(x, y, debug=True) for x in range(self.w)]) for y in range(self.h)]))

        
    # -- about cell informations
    
    def is_cell(x, y):
        return (x, y) in self.cells
        
    def cell(self, x, y):
        return self.cells[x,y]

    def is_wall(self, x, y):
        return self.cell(x, y) == WALL
    
    def is_exit(self, x, y):
        return self.cell(x, y) == EXIT
    
    def inside(self, x, y):
        return (0 < x < self.w-1) and (0 < y < self.h-1)
    
    def reachable(self, x, y):
        return self.is_exit(x, y) or self.inside(x, y) and not self.is_wall(x, y)
    
    # -- about bot informations and updates
    
    def bot_at(self, x, y):
        return (x, y) in self.bots
    
    def bot_by_id(self, bot_id):
        for bot in self.all_bots():
            if bot.id == bot_id:
                return bot
    
    def bot(self, x, y):
        return self.bots[x, y]

    def all_bots(self):
        return self.bots.values()
    
    def remove_bot(self, coords):
        del self.bots[coords]
    
    def let_bot_exit(self):
        if self.exit_coords in self.bots:
            self.remove_bot(self.exit_coords)
    
    
    # -- manage blocked bots
        
    def _propagate_update(self):
        a_traiter = [(x, y) for x, y in self.bots if self.bot(x,y).blocked]
        seen = set()
        while a_traiter:
            x, y = a_traiter.pop(0)
            seen.add((x,y))
            if (x, y) in self.destinations:
                for nx, ny in self.destinations[x,y]:
                    if (nx, ny) not in a_traiter and (nx, ny) not in seen and (nx, ny) in self.destinations:
                        a_traiter.append((nx, ny))
                    self.bot(nx, ny).blocked = True
        
    def _update_destinations(self):
        self.destinations.clear()
        for x, y in self.bots:
            bot = self.bot(x, y)
            dest_x, dest_y = bot.destination
            self.destinations[dest_x, dest_y] = self.destinations.get((dest_x, dest_y), []) + [(x, y)]

    def _update_blocked(self):
        for x, y in self.bots:
            bot = self.bot(x, y)
            dest_x, dest_y = bot.destination
            if not self.reachable(dest_x, dest_y) or len(self.destinations[dest_x, dest_y]) > 1:
                bot.blocked = True

    def update(self):
        self._update_destinations()
        self._update_blocked()
        self._propagate_update()
        self.updated = True
                
    def deadlock(self):
        if not self.updated:
            self.update()
        return all(b.blocked == True for b in self.all_bots())
    
    
    # -- about moving
    
    def left(self):
        self.update()
        for b in self.all_bots():
            b.left()
        self.updated = False

    def right(self):
        self.update()
        for b in self.all_bots():
            b.right()
        self.updated = False

    def forward(self):
        self.update()
        copie_bots = {}
        bots_coords = list(self.bots)
        for x, y in bots_coords:
            bot = self.bot(x,y)
            bot.forward()
            copie_bots[bot.position] = bot
        self.bots = copie_bots
        self.let_bot_exit()
        self.updated = False
    
    # -- about solution
    
    def win(self):
        return self.bots == {}
    
    def lose(self):
        count = 0
        if not self.deadlock():
            return False
        for _ in range(3):
            self.left()
            count += 1
            if not self.deadlock():
                for _ in range(count):
                    self.right()
                self.update()
                return False
        self.left()
        self.update()
        return True
    
    def read(self, actions):
        translation = {'G':self.left, 'D':self.right, 'A':self.forward}
        actions = actions.upper()
        for action in actions:
            move = translation[action]
            move()
            clear_output(wait=True)
            print(self)
            sleep(0.5)
        if self.win():
            print('You win')
        elif self.lose():
            print('You lose')
        else:
            print('Not finished')
    